/*
 * pajs_mixer.c

 *
 *  Created on: 10.07.2015
 *      Author: klausj
 */
#include "ips_audio_pulse_PulseMixerProvider.h"
#include "ips_audio_pulse_PulseTargetDataline.h"

#define INITIAL_BUFSIZE 2048
#define MAX_ERR_MSG 2048

JNIEXPORT jobject JNICALL Java_ips_audio_pulse_PulseTargetDataLine_init(JNIEnv *env, jobject jobj,jobject mpNp) {
	PulseMixerProvider* pmp = (*env)->GetDirectBufferAddress(env,mpNp);
	jobject np = NULL;

	JavaVM *javaVM;
	jint res = (*env)->GetJavaVM(env, &javaVM);
	if (res == 0) {

		PulseTargetDataLine *ptdl=(PulseTargetDataLine*)malloc(sizeof(PulseTargetDataLine));
		ptdl->pmp=pmp;
		ptdl->nBufSize=INITIAL_BUFSIZE;
		ptdl->nBuf=malloc(ptdl->nBufSize);

		//ptdl->paCtx=pmp->paCtx;
		//ptdl->m=pmp->paMainloop;
		ptdl->javaVM = javaVM;
		ptdl->javaObjRef = (*env)->NewWeakGlobalRef(env, jobj);

		np= (*env)->NewDirectByteBuffer(env, ptdl,sizeof(PulseTargetDataLine));
		jclass cls = (*env)->GetObjectClass(env, ptdl->javaObjRef);
		if (cls == NULL) {
			fprintf(stderr,"Could not find class\n");
			return NULL;
		}
		ptdl->jClassRef=(*env)->NewGlobalRef(env,cls);
	}
	return(np);
}

// Java : open(nativePointer,name,(int)af.getSampleRate(),af.getSampleSizeInBits(),sampleSize,chs,af.isBigEndian());
JNIEXPORT void JNICALL Java_ips_audio_pulse_PulseTargetDataLine_open(JNIEnv *env, jobject obj,jobject np,jstring jsourceName,jint sampleRate, jint validSampleSizeInBits,jint sampleSizeInBits, jint channels, jboolean bigEndian,jint paBufFragSize,jint bufferSize){

	const char *streamName="PulseAudio JavaSound";
	pa_sample_spec  	ss;
	pa_sample_spec_init(&ss);

	ss.channels=channels;
	ss.format=PA_SAMPLE_INVALID;
	if(validSampleSizeInBits==8){
		ss.format=PA_SAMPLE_U8;
	}else if(validSampleSizeInBits==16){

		if(bigEndian){
			ss.format=PA_SAMPLE_S16BE;
		}else{
			ss.format=PA_SAMPLE_S16LE;
		}

	}else if(validSampleSizeInBits==24){
		if(sampleSizeInBits==24){

			if(bigEndian){
				ss.format=PA_SAMPLE_S24BE;
			}else{
				ss.format=PA_SAMPLE_S24LE;
			}
		}
		if(sampleSizeInBits==32){
			if(bigEndian){
				ss.format=PA_SAMPLE_S24_32BE;
			}else{
				ss.format=PA_SAMPLE_S24_32LE;
			}
		}
	}else if(validSampleSizeInBits==32){
		if(bigEndian){
			ss.format=PA_SAMPLE_S32BE;
		}else{
			ss.format=PA_SAMPLE_S32LE;
		}
	}
	PulseTargetDataLine* ptdl = (*env)->GetDirectBufferAddress(env,np);
	if(ss.format==PA_SAMPLE_INVALID){
		(*env)->ThrowNew(env,ptdl->pmp->exceptionClassRef,"Could not get object class");
		fprintf(stderr,"Could not get object class");
		return;
	}
#ifdef DEBUG
	printf("PA: record sample format %s\n",pa_sample_format_to_string(ss.format));
	fflush(stdout);
#endif
	ss.rate=sampleRate;

	int res=pa_sample_spec_valid(&ss);
	if(res<=0){
		char errMsg[MAX_ERR_MSG+1];
		snprintf(errMsg,MAX_ERR_MSG,"PulseAudio: sample spec invalid: %s",pa_strerror(res));
		fprintf(stderr,"%s\n",errMsg);
		fflush(stderr);
		(*env)->ThrowNew(env,ptdl->pmp->exceptionClassRef,errMsg);
	}
	if(bufferSize>0){
		if(ptdl->nBuf){
			free(ptdl->nBuf);
		}
		ptdl->nBuf=malloc(bufferSize);
	}
	ptdl->nBufAvail=0;
	ptdl->nBufPos=0;
	ptdl->paStream=pa_stream_new (ptdl->pmp->paCtx,streamName,&ss,NULL);
	pa_stream_set_state_callback(ptdl->paStream,rec_stream_cb,ptdl);
	jboolean isCopy;
	const char *sourceName = (*env)->GetStringUTFChars(env,jsourceName, &isCopy);

	//pa_buffer_attr buffAttr;

	ptdl->buffAttr.fragsize=paBufFragSize;
	ptdl->buffAttr.maxlength=-1;
	ptdl->buffAttr.minreq=-1;
	ptdl->buffAttr.tlength=-1;
	ptdl->buffAttr.prebuf=-1;

	pa_stream_flags_t flags=PA_STREAM_ADJUST_LATENCY;

	pa_threaded_mainloop_lock(ptdl->pmp->paMainloop);
	res=pa_stream_connect_record(ptdl->paStream,sourceName,
			&(ptdl->buffAttr),
			flags
	);
	if(res!=0){
		char errMsg[MAX_ERR_MSG+1];
		snprintf(errMsg,MAX_ERR_MSG,"PulseAudio: Record connect failed: %s",pa_strerror(res));
		fprintf(stderr,"%s\n",errMsg);
		fflush(stderr);
		(*env)->ThrowNew(env,ptdl->pmp->exceptionClassRef,errMsg);
		return;
	}
	pa_threaded_mainloop_unlock(ptdl->pmp->paMainloop);

	(*env)->ReleaseStringUTFChars(env,jsourceName,sourceName);

}

JNIEXPORT jint JNICALL Java_ips_audio_pulse_PulseTargetDataLine_available(JNIEnv *env, jobject obj,jobject np) {
	PulseTargetDataLine* ptdl = (*env)->GetDirectBufferAddress(env,np);
	jint av=ptdl->nBufAvail;
	if(av<=0){
		pa_stream* stream=ptdl->paStream;
		//pa_threaded_mainloop_lock(ptdl->m);

		// some times returns -1 :
		// http://pulseaudio-discuss.freedesktop.narkive.com/YZIkSMSq/pa-stream-readable-size-returns-4294967295
		// if stream is not in READY state !!
		av=pa_stream_readable_size(stream);
		//pa_threaded_mainloop_unlock(ptdl->m);
	}
	return av;
}

JNIEXPORT jint JNICALL Java_ips_audio_pulse_PulseTargetDataLine_read(JNIEnv* env, jobject obj,jobject np,jbyteArray buf,jint offset,jint len){

	PulseTargetDataLine* ptdl = (*env)->GetDirectBufferAddress(env,np);
	if(ptdl->nBufAvail<=0){
		// fill native buffer
		size_t peeked;
		const void *data;

		pa_threaded_mainloop_lock(ptdl->pmp->paMainloop);

		int res=pa_stream_peek(ptdl->paStream,&data,&peeked);
		if(res){
			pa_threaded_mainloop_unlock(ptdl->pmp->paMainloop);
			char errMsg[MAX_ERR_MSG+1];
			snprintf(errMsg,MAX_ERR_MSG,"PulseAudio: Stream peek failed: %s",pa_strerror(res));
			fprintf(stderr,"%s\n",errMsg);
			fflush(stderr);
			(*env)->ThrowNew(env,(*env)->FindClass(env,"java/io/IOException"),pa_strerror(res));
			return 0;
		}

		if(ptdl->nBufSize<peeked){
			// dynamically increasing buffer
			free((void *)ptdl->nBuf);
			ptdl->nBuf=malloc(peeked);
			ptdl->nBufPos=0;
			ptdl->nBufSize=peeked;
		}
		if(data!=NULL && peeked>0){
			ptdl->nBufPos=0;
			memcpy((void *)ptdl->nBuf,data,peeked);

			ptdl->nBufAvail=peeked;
		}
		pa_stream_drop(ptdl->paStream);
		pa_threaded_mainloop_unlock(ptdl->pmp->paMainloop);
#ifdef DEBUG
		printf("PA: peeked %zd bytes\n",peeked);
		fflush(stdout);
#endif
	}
	jint toRead=len;
	if(toRead>ptdl->nBufAvail){
		toRead=ptdl->nBufAvail;
	}

	(*env)->SetByteArrayRegion(env,buf,offset,toRead,ptdl->nBuf+ptdl->nBufPos);
	ptdl->nBufAvail-=toRead;
	ptdl->nBufPos+=toRead;
	return toRead;
}

void rec_stream_cb(pa_stream *s,void *userdata){


	PulseTargetDataLine* psdl = userdata;
	pa_stream_state_t sst=pa_stream_get_state 	(psdl->paStream);
	char* mNm=NULL;
	if(sst==PA_STREAM_FAILED){
		mNm="streamFailed";
	}else if(sst==PA_STREAM_READY){
		mNm="streamReady";
	}if(sst==PA_STREAM_TERMINATED){
		mNm="streamTerminated";
	}
	if(mNm!=NULL){
		JNIEnv *env;
		JavaVM *jvm = psdl->javaVM;
		if (jvm != NULL) {

			jint res = (*jvm)->AttachCurrentThread(jvm, (void **) &env, NULL);
			if (res == 0) {
				jclass cls = (*env)->GetObjectClass(env, psdl->javaObjRef);
				if (cls == NULL) {
					printf("Could not find class\n");
					return;
				}
				jmethodID mId=NULL;

				mId=(*env)->GetMethodID(env,cls,mNm,"()V");
				if(mId==NULL){
					fprintf(stderr,"Method not found!");
					return;
				}
				(*env)->CallVoidMethod(env,psdl->javaObjRef,mId);
				//printf("PA: stream callback java called\n");
				//fflush(stdout);
				(*jvm)->DetachCurrentThread(jvm);
			}
		}
	}
}

void flush_ptdl_cb(pa_stream *s, int success, void *userdata){

	PulseTargetDataLine* ptdl = userdata;
	JNIEnv *env;
	JavaVM *jvm = ptdl->javaVM;
	if (jvm != NULL) {
		jint res = (*jvm)->AttachCurrentThread(jvm, (void **) &env, NULL);
		if (res == 0) {
			jclass cls = (*env)->GetObjectClass(env, ptdl->javaObjRef);
			if (cls == NULL) {
				fprintf(stderr,"Could not find class\n");
				return;
			}
			jmethodID mId=NULL;

			mId=(*env)->GetMethodID(env,cls,"flushed","(I)V");
			if(mId==NULL){
				fprintf(stderr,"Method not found!");
				return;
			}
			(*env)->CallVoidMethod(env,ptdl->javaObjRef,mId,success);
			(*jvm)->DetachCurrentThread(jvm);
		}
	}
}

JNIEXPORT void JNICALL Java_ips_audio_pulse_PulseTargetDataLine_flushBegin(JNIEnv *env, jobject obj,jobject np) {
	PulseTargetDataLine* ptdl = (*env)->GetDirectBufferAddress(env,np);
	pa_stream* stream=ptdl->paStream;
	pa_threaded_mainloop_lock(ptdl->pmp->paMainloop);
	pa_stream_flush(stream,flush_ptdl_cb,ptdl);
	pa_threaded_mainloop_unlock(ptdl->pmp->paMainloop);
}

JNIEXPORT jlong JNICALL Java_ips_audio_pulse_PulseTargetDataLine_bytePosition(JNIEnv *env, jobject obj,jobject np) {
	PulseTargetDataLine* ptdl = (*env)->GetDirectBufferAddress(env,np);
	pa_stream* stream=ptdl->paStream;
	jlong bytePos=0;
	pa_threaded_mainloop_lock(ptdl->pmp->paMainloop);
	// TODO dirty, not asynchronous
	pa_stream_update_timing_info(stream,NULL,ptdl);
	const pa_timing_info* ti=pa_stream_get_timing_info(stream);
	if(ti!=NULL){
		bytePos=ti->read_index;
	}
	pa_threaded_mainloop_unlock(ptdl->pmp->paMainloop);

	return bytePos;
}

JNIEXPORT jlong JNICALL Java_ips_audio_pulse_PulseTargetDataLine_usecPosition(JNIEnv *env, jobject obj,jobject np) {
	PulseTargetDataLine* ptdl = (*env)->GetDirectBufferAddress(env,np);
	pa_stream* stream=ptdl->paStream;
	pa_usec_t usecPos=0;
	int res=pa_stream_get_time(stream,&usecPos);
	if(!res){
		return 0;
	}
	return usecPos;
}

JNIEXPORT void JNICALL Java_ips_audio_pulse_PulseTargetDataLine_close(JNIEnv *env, jobject obj,jobject np) {
	PulseTargetDataLine *ptdl = (*env)->GetDirectBufferAddress(env,np);
	pa_threaded_mainloop_lock(ptdl->pmp->paMainloop);
	int res=pa_stream_disconnect(ptdl->paStream);
	pa_threaded_mainloop_unlock(ptdl->pmp->paMainloop);
	if(res){
		fprintf(stderr,"PulseAudio: disconnect: %s\n",pa_strerror(res));
		fflush(stderr);
	}

}

JNIEXPORT void JNICALL Java_ips_audio_pulse_PulseTargetDataLine_release(JNIEnv *env, jobject obj,jobject np) {
	PulseTargetDataLine *ptdl = (*env)->GetDirectBufferAddress(env,np);
	pa_stream_unref(ptdl->paStream);
}




